;---------------------------------------------------------------------------
;
; 		   32-bit Dynamic Linking API Routines
;
;     Copyright (c) 1999-2001 by Thomas Kjoernes, all rights reserved.
;
;
;	API Routines:
;
;	- KERNEL_ApiRegister
;	- KERNEL_ApiUnRegister
;	- KERNEL_ApiGetProcEntry
;	- KERNEL_ApiGetProcAddress
;
;---------------------------------------------------------------------------

	INCLUDE	std.inc
	INCLUDE	stdgroup.inc
	INCLUDE	kernel.api

	.386p

	; Exports

	PUBLIC	apiRegister
	PUBLIC	apiUnRegister
	PUBLIC	apiGetProcEntry
	PUBLIC	apiGetProcAddress

	; Startup/Shutdown

	PUBLIC	apiInit
	PUBLIC	apiDone

codeSegment SEGMENT

	EXTERN	idtSetEntry:NEAR

;---------------------------------------------------------------------------
;	apiInt40Handler - Ring0 Dynamic Link Handler
;
;	This routine services interrupt 40h calls made by kernel-mode
;	device drivers, service providers and any other Ring0 clients.
;
;	The APIcall and APIjmp macros encode INT 40h followed by a DWORD
;	that describes the service to call. Assumtions are made inside
;	this handler about DS being the standard DATA32 selector. 
;
;	Entry (Stack):
;
;		DWORD	EIP
;		DWORD	CS
;		DWORD	EFLAGS
;	Exit:
;		None
;
DefProc apiInt40Handler

	push	eax
	push	edx
	push	esi
	push	edi

	mov	esi, [esp+16]		; get return EIP
	mov	eax, [esi]		; get service ID

	add	esi, 4			; new return address

	mov	[esp+16], esi		; and update stack

	call	apiGetProcEntry		; get procedure entry
	jc	apiInt40HndFail		;

	add	edi, 4			; edi -> [apiProcAddr]

	mov	edx, [edi]	   	; get procedure address
	mov	apiTempProc, edx	; and store it temporarily

	mov	edx, 25FFh		; grp5 -- near indirect jump

	test	ah, ah		  	; call or jmp?
	js	apiInt40HndSave		;

	mov	dh, 15h			; grp5 -- near indirect call

apiInt40HndSave:
	mov	[esi-6], dx		; save opcode
	mov	[esi-4], edi		; save memory address

apiInt40HndDone:
	test	ah, ah			; call or jmp?

	pop	edi
	pop	esi
	pop	edx
	pop	eax

	pop	apiTempEIP		; save return EIP
	pop	apiTempCS		; save return CS
	pop	apiTempEFLAGS		; save caller flags

	js	apiInt40HndJump

	push	apiTempEIP		; push return EIP for service

apiInt40HndJump:

	; stack contents for "IRETD"

	push	apiTempEFLAGS		; push caller flags
	push	apiTempCS		; push caller CS
	push	apiTempProc		; push the service address

	push	EFLAGS_TF		; debugger support
	popfd				;

apiInt40HndExit:
	iretd				; call [apiProcAddr]

apiInt40HndFail:
	mov	apiTempProc, OFFSET apiInt40HndStub
	jmp	apiInt40HndDone

apiInt40HndStub:
	stc
	ret

EndProc apiInt40Handler

;---------------------------------------------------------------------------
;	apiRegister
;
;	Entry:
;		esi = pointer to API_TABLE structure
;	Exit:
;		None
;
DefProc apiRegister

	push	eax
	push	ebx

	lea	eax, [esi].apiListItem
	lea	ebx, apiListHead
	APIcall	KERNEL_ListAttachTail

	pop	ebx
	pop	eax
	ret

EndProc apiRegister

;---------------------------------------------------------------------------
;	apiUnRegister
;
;	Entry:
;		esi = pointer to API_TABLE structure
;	Exit:
;		None
;
DefProc apiUnRegister

	push	eax
	push	ebx

	lea	eax, [esi].apiListItem
	lea	ebx, apiListHead
	APIcall	KERNEL_ListDelete

	pop	ebx
	pop	eax
	ret

EndProc apiUnRegister

;---------------------------------------------------------------------------
;	apiGetProcEntry
;
;	Entry:
;		eax = service ID
;	Exit:
;		edi = pointer to service entry (API_ENTRY structure)
;
DefProc apiGetProcEntry

	push	eax
	push	ebx
	push	ecx
	push	edx

	and	eax, 7FFF7FFFh
	mov	ebx, apiListHead.listHead

apiGetProcEntryMain:
	mov	edi, [ebx].apiProcTable
	mov	ecx, [ebx].apiProcCount
	jecxz	apiGetProcEntrySkip

apiGetProcEntryNext:
	mov	edx, [edi].apiProcID
	and	edx, 7FFF7FFFh

	cmp	eax, edx
	je	apiGetProcEntryExit

	add	edi, SIZE API_ENTRY
	dec	ecx
	jnz	apiGetProcEntryNext

apiGetProcEntrySkip:
	mov	ebx, [ebx].apiListItem.listNext
	test	ebx, ebx
	jnz	apiGetProcEntryMain
 
apiGetProcEntryFail:
	xor	edi, edi
	stc

apiGetProcEntryExit:
	pop	edx
	pop	ecx
	pop	ebx
	pop	eax
	ret

EndProc apiGetProcEntry

;---------------------------------------------------------------------------
;	apiGetProcAddress
;
;	Entry:
;		eax = service ID
;	Exit:
;		ebx = pointer to service handler
;
DefProc apiGetProcAddress

	push	edi

	call	apiGetProcEntry
	jc	apiGetProcAddressExit

	mov	ebx, [edi].apiProcAddr

apiGetProcAddressExit:
	pop	edi
	ret

EndProc apiGetProcAddress

;---------------------------------------------------------------------------
;	apiInit
;
;	Entry:
;		esi = Pointer to KERNEL32 API Table
;	Exit:
;		None
;
DefProc apiInit

	pushad

	; DPL=0 Interrupt gate

	mov	eax, 40h
	mov	ebx, OFFSET apiInt40Handler
	mov	ecx, cs
	mov	edx, DESC_INTR_GATE
	call	idtSetEntry

	popad

	ret

EndProc apiInit

;---------------------------------------------------------------------------
;	apiDone
;
;	Entry:
;		None
;	Exit:
;		None
;
DefProc apiDone
	ret
EndProc apiDone

codeSegment ENDS

;---------------------------------------------------------------------------
; Data
;---------------------------------------------------------------------------

dataSegment SEGMENT

	EXTERN	apiServiceTbl

dataSegment ENDS

;---------------------------------------------------------------------------
; BSS Data
;---------------------------------------------------------------------------

bssSegment SEGMENT

apiListHead	LIST_HEAD <apiServiceTbl, apiServiceTbl>

apiTempProc	DD	?
apiTempCS	DD	?
apiTempEIP	DD	?
apiTempEFLAGS	DD	?

bssSegment ENDS

	END
